import random
import pprint
import time
import uuid
import tempfile
import os, sys
from copy import copy
from socket import gethostname
import cloudpickle as pickle

import numpy as np

import absl.flags
from absl import logging
from ml_collections import ConfigDict
from ml_collections.config_flags import config_flags
from ml_collections.config_dict import config_dict

import wandb
from wandb_osh.hooks import TriggerWandbSyncHook  

from .jax_utils import init_rng


def check_cql_if_run_exists(project, configs):
  keys_of_interests = [
    "env",
    "seed",
    "sampler",
    "reward_norm",
    "cql.discriminator_clip_ratio",
    "cql.discriminator_kl_penalty",
    "cql.discriminator_lr",
    "cql.discriminator_temp",
    "cql.cql_min_q_weight",
    "policy_arch",
  ]

  # Check if the specified configuration has been run
  prefixed_configs = {f"config.{key_of_interest}": configs[key_of_interest]
    for key_of_interest in keys_of_interests if not key_of_interest.startswith("cql")}
  api = wandb.Api()
  old_runs = list(api.runs(path=f"{entity}/{project}",
          filters={
              "state": {"$in": ["finished", "running"]},
              **prefixed_configs,
          }))

  is_exist = np.any([
    np.all([old_run.config[key_of_interest] == configs[key_of_interest] for key_of_interest in keys_of_interests]) for old_run in old_runs
  ])

  return is_exist


class Timer(object):

    def __init__(self):
        self._time = None

    def __enter__(self):
        self._start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._time = time.time() - self._start_time

    def __call__(self):
        return self._time


class WandBLogger(object):

    @staticmethod
    def get_default_config(updates=None):
        config = ConfigDict()
        config.online = False
        config.prefix = 'JaxCQL'
        config.project = ''
        config.output_dir = '/tmp/JaxCQL'
        config.random_delay = 0.0
        config.experiment_id = config_dict.placeholder(str)
        config.anonymous = config_dict.placeholder(str)
        config.notes = config_dict.placeholder(str)
        config.wandb_api_key = config_dict.placeholder(str) # None: you login in shell

        if updates is not None:
            config.update(ConfigDict(updates).copy_and_resolve_references())
        return config

    def __init__(self, config, variant):
        self.config = self.get_default_config(config)

        if self.config.experiment_id is None:
            self.config.experiment_id = uuid.uuid4().hex

        if self.config.prefix != '':
            self.config.project = '{}{}'.format(self.config.prefix, self.config.project)

        if self.config.output_dir == '':
            self.config.output_dir = tempfile.mkdtemp()
        else:
            self.config.output_dir = os.path.join(self.config.output_dir, self.config.experiment_id)
            os.makedirs(self.config.output_dir, exist_ok=True)

        self._variant = copy(variant)

        if 'hostname' not in self._variant:
            self._variant['hostname'] = gethostname()

        if self.config.random_delay > 0:
            time.sleep(np.random.uniform(0, self.config.random_delay))

        if self.config.wandb_api_key != config_dict.placeholder(str):
          wandb.login(key=self.config.wandb_api_key)

        # Check redundant runs
        if os.environ.get("DEBUG", None):
            self.check_redundant_runs()
        
        if not self.config.online:
            self.trigger_sync = TriggerWandbSyncHook()

        self.run = wandb.init(
            reinit=True,
            config=self._variant,
            project=self.config.project,
            dir=self.config.output_dir,
            id=self.config.experiment_id,
            anonymous=self.config.anonymous,
            notes=self.config.notes,
            settings=wandb.Settings(
                start_method="thread",
                _disable_stats=True,
            ),
            mode='online' if self.config.online else 'offline',            
        )

    def check_redundant_runs(self):
        if check_cql_if_run_exists(self._variant,
                project=self.config.project):
          print("Run exists")
          sys.exit()

    def log(self, *args, **kwargs):
        self.run.log(*args, **kwargs)
        if not self.config.online:
            self.trigger_sync()

    def save_pickle(self, obj, filename):
        with open(os.path.join(self.config.output_dir, filename), 'wb') as fout:
            pickle.dump(obj, fout)

    @property
    def experiment_id(self):
        return self.config.experiment_id

    @property
    def variant(self):
        return self.config.variant

    @property
    def output_dir(self):
        return self.config.output_dir


def define_flags_with_default(**kwargs):
    for key, val in kwargs.items():
        if isinstance(val, ConfigDict):
            config_flags.DEFINE_config_dict(key, val)
        elif isinstance(val, bool):
            # Note that True and False are instances of int.
            absl.flags.DEFINE_bool(key, val, 'automatically defined flag')
        elif isinstance(val, int):
            absl.flags.DEFINE_integer(key, val, 'automatically defined flag')
        elif isinstance(val, float):
            absl.flags.DEFINE_float(key, val, 'automatically defined flag')
        elif isinstance(val, str):
            absl.flags.DEFINE_string(key, val, 'automatically defined flag')
        elif isinstance(val, list):
            absl.flags.DEFINE_list(key, val, 'automatically defined flag')
        else:
            raise ValueError('Incorrect value type')
    return kwargs


def set_random_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    init_rng(seed)


def print_flags(flags, flags_def):
    logging.info(
        'Running training with hyperparameters: \n{}'.format(
            pprint.pformat(
                ['{}: {}'.format(key, val) for key, val in get_user_flags(flags, flags_def).items()]
            )
        )
    )


def get_user_flags(flags, flags_def):
    output = {}
    for key in flags_def:
        val = getattr(flags, key)
        if isinstance(val, ConfigDict):
            output.update(flatten_config_dict(val, prefix=key))
        else:
            output[key] = val

    return output


def flatten_config_dict(config, prefix=None):
    output = {}
    for key, val in config.items():
        if prefix is not None:
            next_prefix = '{}.{}'.format(prefix, key)
        else:
            next_prefix = key
        if isinstance(val, ConfigDict):
            output.update(flatten_config_dict(val, prefix=next_prefix))
        else:
            output[next_prefix] = val
    return output



def prefix_metrics(metrics, prefix):
    return {
        '{}/{}'.format(prefix, key): value for key, value in metrics.items()
    }
